@@ -32,6 +32,8 @@
 #include <QJsonObject>
 #include <QMutex>
 #include <QStandardPaths>
+#include <QRegularExpression>
+#include <QRegularExpressionMatch>
 
 // in the desktoptojson binary enable debug messages by default, in the library only warning messages
 #ifdef BUILDING_DESKTOPTOJSON_TOOL
@@ -226,23 +228,83 @@ QByteArray readTypeEntryForCurrentGroup(QFile &df, QByteArray *nextGroup)
             *nextGroup = name;
             break;
         }
-        if (line.startsWith(QByteArrayLiteral("Type="))) {
-            // TODO: should we also have to accept spaces around equals here?
-            type = line.mid(qstrlen("Type="));
+
+        const static QRegularExpression typeEntryRegex(
+                QStringLiteral("^Type\\s*=\\s*(.*)$"));
+        const auto match = typeEntryRegex.match(QString::fromUtf8(line));
+        if (match.hasMatch()) {
+            type = match.captured(1).toUtf8();
         }
     }
     return type;
 }
 
-QVector<CustomPropertyDefinition>* parseServiceTypesFile(QString path)
+bool tokenizeKeyValue(QFile &df, const QString &src, QByteArray &key, QString &value, int &lineNr)
+{
+    const QByteArray line = df.readLine().trimmed();
+    lineNr++;
+    if (line.isEmpty()) {
+        DESKTOPTOJSON_VERBOSE_DEBUG << "Line " << lineNr << ": empty";
+        return true;
+    }
+    if (line.startsWith('#')) {
+        DESKTOPTOJSON_VERBOSE_DEBUG << "Line " << lineNr << ": comment";
+        return true; // skip comments
+    }
+    if (line.startsWith('[')) {
+        // start of new group -> doesn't interest us anymore
+        DESKTOPTOJSON_VERBOSE_DEBUG << "Line " << lineNr << ": start of new group " << line;
+        return false;
+    }
+    // must have form key=value now
+    const int equalsIndex = line.indexOf('=');
+    if (equalsIndex == -1) {
+        qCWarning(DESKTOPPARSER).nospace() << qPrintable(src) << ':' << lineNr << ": Line is neither comment nor group "
+            "and doesn't contain an '=' character: \"" << line.constData() << '\"';
+        return true;
+    }
+    // trim key and value to remove spaces around the '=' char
+    key = line.mid(0, equalsIndex).trimmed();
+    if (key.isEmpty()) {
+        qCWarning(DESKTOPPARSER).nospace() << qPrintable(src) << ':' << lineNr << ": Key name is missing: \"" << line.constData() << '\"';
+        return true;
+    }
+
+    const QByteArray valueRaw = line.mid(equalsIndex + 1).trimmed();
+    const QByteArray valueEscaped = escapeValue(valueRaw);
+    value = QString::fromUtf8(valueEscaped);
+
+#ifdef BUILDING_DESKTOPTOJSON_TOOL
+    DESKTOPTOJSON_VERBOSE_DEBUG.nospace() << "Line " << lineNr << ": key=" << key << ", value=" << value;
+    if (valueEscaped != valueRaw) {
+        DESKTOPTOJSON_VERBOSE_DEBUG << "Line " << lineNr << " contained escape sequences";
+    }
+#endif
+
+    return true;
+}
+
+static QString locateRelativeServiceType(const QString &relPath)
+{
+    return QStandardPaths::locate(QStandardPaths::GenericDataLocation,
+                                  QStringLiteral("kservicetypes5/") + relPath);
+}
+
+static QVector<CustomPropertyDefinition>* parseServiceTypesFile(const QString &inputPath)
 {
     int lineNr = 0;
+    QString path = inputPath;
     if (QDir::isRelativePath(path)) {
-        const QString originalPath = path;
-        path = QStandardPaths::locate(QStandardPaths::GenericDataLocation,
-                               QStringLiteral("kservicetypes5/") + path);
+        path = locateRelativeServiceType(path);
+        QString rcPath;
         if (path.isEmpty()) {
-            qCCritical(DESKTOPPARSER) << "Could not locate service type file kservicetypes5/" << originalPath << ", tried" << QStandardPaths::standardLocations(QStandardPaths::GenericDataLocation);
+            rcPath = QStringLiteral(":/kservicetypes5/") + inputPath;
+            if (QFileInfo::exists(rcPath)) {
+                path = rcPath;
+            }
+        }
+        if (path.isEmpty()) {
+            qCWarning(DESKTOPPARSER).nospace() << "Could not locate service type file kservicetypes5/" << qPrintable(inputPath) << ", tried " << QStandardPaths::standardLocations(QStandardPaths::GenericDataLocation) << " and " << rcPath;
             return nullptr;
         }
     }
@@ -307,43 +369,47 @@ QBasicMutex s_serviceTypesMutex;
 } // end of anonymous namespace
 
 
-ServiceTypeDefinition::ServiceTypeDefinition(const QVector< CustomPropertyDefinition >& defs)
-    : m_definitions(defs)
+ServiceTypeDefinition::ServiceTypeDefinition()
 {
 }
 
 ServiceTypeDefinition ServiceTypeDefinition::fromFiles(const QStringList &paths)
 {
-    QVector<CustomPropertyDefinition> defs;
-    defs.reserve(paths.size());
-
+    ServiceTypeDefinition ret;
+    ret.m_definitions.reserve(paths.size());
     // as we might modify the cache we need to acquire a mutex here
-    foreach (const QString &serviceType, paths) {
-        QMutexLocker lock(&s_serviceTypesMutex);
-
-        QVector<CustomPropertyDefinition>* def = s_serviceTypes->object(serviceType);
-
-        if (def) {
-            // in cache but we still must make our own copy
-            defs << *def;
-        }
-        else {
-            // not found in cache -> we need to parse the file
-            qCDebug(DESKTOPPARSER) << "About to parse service type file" << serviceType;
-            def = parseServiceTypesFile(serviceType);
-            if (!def) {
+    foreach (const QString &serviceTypePath, paths) {
+        bool added = ret.addFile(serviceTypePath);
+        if (!added) {
 #ifdef BUILDING_DESKTOPTOJSON_TOOL
-                exit(1); // this is a fatal error when using kcoreaddons_desktop_to_json()
-#else
-                continue;
+            exit(1); // this is a fatal error when using kcoreaddons_desktop_to_json()
 #endif
-            }
-
-            defs << *def; // This must *precede* insert call, insert might delete
-            s_serviceTypes->insert(serviceType, def);
         }
     }
-    return ServiceTypeDefinition(defs);
+    return ret;
+}
+
+bool ServiceTypeDefinition::addFile(const QString& path)
+{
+    QMutexLocker lock(&s_serviceTypesMutex);
+    QVector<CustomPropertyDefinition>* def = s_serviceTypes->object(path);
+
+    if (def) {
+        // in cache but we still must make our own copy
+        m_definitions << *def;
+    }
+    else {
+        // not found in cache -> we need to parse the file
+        qCDebug(DESKTOPPARSER) << "About to parse service type file" << path;
+        def = parseServiceTypesFile(path);
+        if (!def) {
+            return false;
+        }
+
+        m_definitions << *def; // This must *precede* insert call, insert might delete
+        s_serviceTypes->insert(path, def);
+    }
+    return true;
 }
 
 QJsonValue ServiceTypeDefinition::parseValue(const QByteArray &key, const QString &value) const
@@ -358,7 +424,7 @@ QJsonValue ServiceTypeDefinition::parseValue(const QByteArray &key, const QStrin
     return QJsonValue(value);
 }
 
-void DesktopFileParser::convertToJson(const QByteArray &key, const ServiceTypeDefinition &serviceTypes, const QString &value,
+void DesktopFileParser::convertToJson(const QByteArray &key, ServiceTypeDefinition &serviceTypes, const QString &value,
                                       QJsonObject &json, QJsonObject &kplugin, int lineNr)
 {
     /* The following keys are recognized (and added to a "KPlugin" object):
@@ -397,8 +463,9 @@ void DesktopFileParser::convertToJson(const QByteArray &key, const ServiceTypeDe
     } else if (key == QByteArrayLiteral("X-KDE-PluginInfo-Depends")) {
         kplugin[QStringLiteral("Dependencies")] = QJsonArray::fromStringList(deserializeList(value));
     } else if (key == QByteArrayLiteral("X-KDE-ServiceTypes") || key == QByteArrayLiteral("ServiceTypes")) {
-        // some .desktop files still use the legacy ServiceTypes= key
-        kplugin[QStringLiteral("ServiceTypes")] = QJsonArray::fromStringList(deserializeList(value));
+        //NOTE: "X-KDE-ServiceTypes" and "ServiceTypes" were already managed in the first parse step, so this second one is almost a noop
+        const auto services = deserializeList(value);
+        kplugin[QStringLiteral("ServiceTypes")] = QJsonArray::fromStringList(services);
     } else if (key == QByteArrayLiteral("MimeType")) {
         // MimeType is a XDG string list and not a KConfig list so we need to use ';' as the separator
         kplugin[QStringLiteral("MimeTypes")] = QJsonArray::fromStringList(deserializeList(value, ';'));
@@ -464,46 +531,51 @@ bool DesktopFileParser::convert(const QString &src, const QStringList &serviceTy
     ServiceTypeDefinition serviceTypeDef = ServiceTypeDefinition::fromFiles(serviceTypes);
     readUntilDesktopEntryGroup(df, src, lineNr);
     DESKTOPTOJSON_VERBOSE_DEBUG << "Found [Desktop Entry] group in line" << lineNr;
+    auto startPos = df.pos();
 
-    //QJsonObject json;
-    QJsonObject kplugin; // the "KPlugin" key of the metadata
+    //parse it a first time to know servicetype
     while (!df.atEnd()) {
-        const QByteArray line = df.readLine().trimmed();
-        lineNr++;
-        if (line.isEmpty()) {
-            DESKTOPTOJSON_VERBOSE_DEBUG << "Line " << lineNr << ": empty";
-            continue;
-        }
-        if (line.startsWith('#')) {
-            DESKTOPTOJSON_VERBOSE_DEBUG << "Line " << lineNr << ": comment";
-            continue; // skip comments
-        }
-        if (line.startsWith('[')) {
-            // start of new group -> doesn't interest us anymore
-            DESKTOPTOJSON_VERBOSE_DEBUG << "Line " << lineNr << ": start of new group " << line;
+        QByteArray key;
+        QString value;
+        if (!tokenizeKeyValue(df, src, key, value, lineNr)) {
             break;
         }
-        // must have form key=value now
-        const int equalsIndex = line.indexOf('=');
-        if (equalsIndex == -1) {
-            qCWarning(DESKTOPPARSER).nospace() << qPrintable(src) << ':' << lineNr << ": Line is neither comment nor group "
-                "and doesn't contain an '=' character: \"" << line.constData() << '\"';
+        if (key == QByteArrayLiteral("X-KDE-ServiceTypes") || key == QByteArrayLiteral("ServiceTypes")) {
+            const QString dotDesktop = QStringLiteral(".desktop");
+            const QChar slashChar(QLatin1Char('/'));
+            const auto serviceList = deserializeList(value);
+
+            for(const auto &service : serviceList) {
+                // some .desktop files still use the legacy ServiceTypes= key
+                QString absFileName = locateRelativeServiceType(
+                        service.toLower().replace(slashChar, QLatin1Char('-'))+dotDesktop);
+                if (absFileName.isEmpty()) {
+                    absFileName = locateRelativeServiceType(
+                        service.toLower().remove(slashChar)+dotDesktop);
+                }
+                if (absFileName.isEmpty()) {
+                    qCWarning(DESKTOPPARSER) << "Unable to find service type for service" << service;
+                }
+                else
+                    serviceTypeDef.addFile(absFileName);
+            }
+            break;
+        }
+    }
+    lineNr=0;
+    df.seek(startPos);
+
+    QJsonObject kplugin; // the "KPlugin" key of the metadata
+    //QJsonObject json;
+    while (!df.atEnd()) {
+        QByteArray key;
+        QString value;
+        if (!tokenizeKeyValue(df, src, key, value, lineNr)) {
+            break;
+        } else if (key.isEmpty()) {
             continue;
         }
-        // trim key and value to remove spaces around the '=' char
-        const QByteArray key = line.mid(0, equalsIndex).trimmed();
-        if (key.isEmpty()) {
-            qCWarning(DESKTOPPARSER).nospace() << qPrintable(src) << ':' << lineNr << ": Key name is missing: \"" << line.constData() << '\"';
-            continue;
-        }
-        const QByteArray valueRaw = line.mid(equalsIndex + 1).trimmed();
-        const QByteArray valueEscaped = escapeValue(valueRaw);
-        const QString value = QString::fromUtf8(valueEscaped);
 #ifdef BUILDING_DESKTOPTOJSON_TOOL
-        DESKTOPTOJSON_VERBOSE_DEBUG.nospace() << "Line " << lineNr << ": key=" << key << ", value=" << value;
-        if (valueEscaped != valueRaw) {
-            DESKTOPTOJSON_VERBOSE_DEBUG << "Line " << lineNr << " contained escape sequences";
-        }
         if (s_compatibilityMode) {
             convertToCompatibilityJson(QString::fromUtf8(key), value, json, lineNr);
         } else {
